
// =====================================================================
//  File:		QueryExpressionHelper.cs
//  Summary:	[FILE DESCRIPTION GOES HERE]
// =====================================================================
//
//  This file is part of the Microsoft CRM 3.0 SDK Code Samples.
//
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  This source code is intended only as a supplement to Microsoft
//  Development Tools and/or on-line documentation.  See these other
//  materials for detailed information regarding Microsoft code samples.
//
//  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//  PARTICULAR PURPOSE.
//
// =====================================================================

using System;
using System.Globalization;
using MsCrmHelpers.CrmSdk;
using Microsoft.Crm.Sdk.Helpers.Collection;

namespace Microsoft.Crm.Sdk.Helpers
{
	public class QueryExpressionHelper
	{
		#region Fields
		private LinkEntityHelperCollection	_linkEntities;
		private FilterExpressionHelper		_criteria;
		private OrderExpressionCollection	_orders;
		private ColumnSetHelper				_columns;

		private QueryExpression				_query;
		#endregion

		#region Properties
		/// <summary>
		/// Specifies the entity that will be queried.
		/// </summary>
		public string EntityName
		{
			get
			{
				return _query.EntityName;
			}
			set
			{
				_query.EntityName = value;
			}
		}

		/// <summary>
		/// Specifies the paging information of how to return the query.
		/// </summary>
		public PagingInfo PageInfo
		{
			get
			{
				// Because the PagingInfo may not be used, auto-instantiation occurs in the "get" property.
				if (_query.PageInfo == null)
				{
					_query.PageInfo = new PagingInfo();
				}

				return _query.PageInfo;
			}
			set
			{
				_query.PageInfo = value;
			}
		}

		/// <summary>
		/// A collection of linked entities that are used to connect the main entity with other entities
		/// for more complex queries.
		/// </summary>
		public LinkEntityHelperCollection LinkEntities
		{
			get
			{
				return _linkEntities;
			}
		}

		/// <summary>
		/// A filter that contains a collection of conditions (and optionally, nested filters) that will be applied to the query.
		/// </summary>
		public FilterExpressionHelper Criteria
		{
			get
			{
				return _criteria;
			}
			set
			{
				_criteria = value;
			}
		}

		/// <summary>
		/// Specifies the order/organization of the query.
		/// </summary>
		public OrderExpressionCollection Orders
		{
			get
			{
				return _orders;
			}
		}

		/// <summary>
		/// Specifies which columns and attributes will be returned by the query
		/// </summary>
		public ColumnSetHelper Columns
		{
			get
			{
				return _columns;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public bool Distinct
		{
			get
			{
				return _query.Distinct;
			}
			set
			{
				_query.Distinct = value;
			}
		}

		/// <summary>
		/// Returns a QueryExpression object.
		/// </summary>
		public QueryExpression Query
		{
			get
			{
				// Setup the QueryExpression prior to returning.
				SetColumns();
				SetOrders();
				SetFilter();
				SetLinkEntities();

				return _query;
			}
		}


		#region Query Property Helper Methods
		/// <summary>
		/// Used in the Query property to convert and set the ColumnBase object of the QueryExpression object.
		/// </summary>
		private void SetColumns()
		{
			// If the user chose all columns
			if (_columns != null && _columns.AllColumns)
			{
				_query.ColumnSet = new AllColumns();
			}
			else if (_columns != null && _columns.Attributes.Count > 0)
			{
				// create a ColumnSet object to assign to avoid multiple casting
				ColumnSet builtinColumnSet = new ColumnSet();
				builtinColumnSet.Attributes = _columns.ToStringArray();

				// re-assign the temporary ColumnSet object to the query
				_query.ColumnSet = builtinColumnSet;
			}
			else
			{
				// Do nothing if the user did not specify any columns.
			}
		}


		/// <summary>
		/// Used in the Query property to convert and set the OrderExpression[] of the QueryExpression object.
		/// </summary>
		private void SetOrders()
		{
			// Checks to make sure that the user specified orders.
			if (_orders != null && _orders.Count > 0)
			{
				_query.Orders = _orders.ToArray();
			}
		}


		/// <summary>
		/// Used in the Query property to convert and set the FilterExpression object of the QueryExpression object.
		/// </summary>
		private void SetFilter()
		{
			// If filters and conditions have been set by the user, use them to construct the actual query.
			if (_criteria != null && _criteria.Conditions.Count > 0)
			{
				_query.Criteria = BuildFilter(_criteria);
			}
			else
			{
				// Do nothing because the user did not specify any filters or conditions.
			}
		}


		/// <summary>
		/// Recursive function that dynamically creates a FilterExpression object from a FilterExpressionHelper object
		/// (linking all the conditions and nested filters).
		/// </summary>
		/// <param name="helperFilter">A FilterExpressionHelper object that needs to be converted.</param>
		/// <returns>A FilterExpression object that can be used in the QueryExpression object.</returns>
		private FilterExpression BuildFilter(FilterExpressionHelper helperFilter)
		{
			FilterExpression builtinFilter	= new FilterExpression();
			builtinFilter.FilterOperator	= helperFilter.FilterOperator;
			builtinFilter.Conditions		= helperFilter.Conditions.ToConditionExpressionArray();

			// If there are nested filters
			if (helperFilter.Filters != null && helperFilter.Filters.Count > 0)
			{
				// recursively iterates through child filters
				FilterExpression[] childFilters = new FilterExpression[helperFilter.Filters.Count];
				for (int index = 0; index < childFilters.Length; index++)
				{                                                    
					childFilters[index] = BuildFilter(helperFilter.Filters[index]);
				}
				builtinFilter.Filters = childFilters;
			}

			return builtinFilter;
		}

		
		/// <summary>
		/// Used in the Query property object to convert and set the LinkEntity[] of the QueryExpression object.
		/// </summary>
		private void SetLinkEntities()
		{
			// If link entities have been set by the user, use them to construct the actual query.
			if (_linkEntities != null && _linkEntities.Count > 0)
			{
				_query.LinkEntities = new LinkEntity[_linkEntities.Count];

				// Iterate through all the linked entities.
				for (int index = 0; index < _query.LinkEntities.Length; index++)
				{
					_query.LinkEntities[index] = BuildLinkEntity(_linkEntities[index]);
				}
			}
			else
			{
				// Do nothing because the user did not specify any linked entities.
			}
		}


		/// <summary>
		/// Recursive function that dynamically creates a LinkEntity object from a LinkEntityHelper object
		/// (linking all the nested linked entities as well as the filter).
		/// </summary>
		/// <param name="helperLinkEntity">A LinkEntityHelper object that needs to be converted.</param>
		/// <returns>A LinkEntity object that can be used in the QueryExpression object.</returns>
		private LinkEntity BuildLinkEntity(LinkEntityHelper helperLinkEntity)
		{
			LinkEntity builtinLinkEntity = new LinkEntity();

			// Sets the properties.
			builtinLinkEntity.LinkFromEntityName	= helperLinkEntity.FromEntityName;
			builtinLinkEntity.LinkFromAttributeName	= helperLinkEntity.FromAttributeName;
			builtinLinkEntity.LinkToEntityName		= helperLinkEntity.ToEntityName;
			builtinLinkEntity.LinkToAttributeName	= helperLinkEntity.ToAttributeName;
			builtinLinkEntity.JoinOperator			= helperLinkEntity.JoinOperator;

			builtinLinkEntity.LinkCriteria			= BuildFilter(helperLinkEntity.Criteria);

			// If there are nested linked entities
			if (helperLinkEntity.LinkEntities != null && helperLinkEntity.LinkEntities.Count > 0)
			{
				// recursively iterates through child linked entities.
				LinkEntity[] childLinkEntities = new LinkEntity[helperLinkEntity.LinkEntities.Count];
				for (int index = 0; index < childLinkEntities.Length; index++)
				{
					childLinkEntities[index] = BuildLinkEntity(helperLinkEntity.LinkEntities[index]);
				}
				builtinLinkEntity.LinkEntities = childLinkEntities;
			}

			return builtinLinkEntity;
		}

		#endregion

		#endregion

		#region Constructor
		public QueryExpressionHelper() {}

		/// <summary>
		/// Used to aid in the creation of a query.
		/// </summary>
		/// <param name="name">An EntityName enum that specifies which entity.</param>
		public QueryExpressionHelper(EntityName name) : this(name.ToString(CultureInfo.InvariantCulture))
		{
		}

		/// <summary>
		/// Used to aid in the creation of a query.
		/// </summary>
		/// <param name="entityName">A string that corresponds to the name of a custom or built-in entity.</param>
		public QueryExpressionHelper(string entityName)
		{
			// Instantiate the QueryExpression object.
			_query				= new QueryExpression();
			_query.EntityName	= entityName;
			
			// Instantiate custom objects.
			_columns			= new ColumnSetHelper();
			_criteria			= new FilterExpressionHelper();
			_orders				= new OrderExpressionCollection();
			_linkEntities		= new LinkEntityHelperCollection(_query.EntityName);
		}

		#endregion
	}
}
