using System;
using System.Collections.Generic;
using System.Text;
using MsCrmHelpers.CrmSdk;

namespace MsCrmHelpers
{
    public class QuickQueries
    {
        public static BusinessEntityCollection Query(CrmService crmService, QueryExpression query)
        {
            // Create the request object.
            RetrieveMultipleRequest retrieve = new RetrieveMultipleRequest();

            // Set the properties of the request object.
            retrieve.Query = query;
            
            // Execute the request.
            RetrieveMultipleResponse searchResults = (RetrieveMultipleResponse)crmService.Execute(retrieve);

            return searchResults.BusinessEntityCollection;
        }

        public static BusinessEntityCollection Query(CrmService service, string entityName)
        {
            return Query(service, BuildQuery(entityName));
        }

        public static QueryExpression BuildQuery(string entityName)
        {
            if (String.IsNullOrEmpty(entityName))
                throw new ArgumentNullException("DataMember");

            // Create the query.
            QueryExpression query = new QueryExpression();

            // Set the query's properties.
            query.EntityName = entityName;
            query.ColumnSet = new AllColumns();

            return query;
        }

        public static BusinessEntityCollection Query(CrmService service, string entityName, string attributeName, object attributeValue)
        {
            return Query(service, BuildQuery( entityName, attributeName, attributeValue));
        }

        public static BusinessEntityCollection Query(CrmService service, EntityName entityName, string attributeName, object attributeValue)
        {
            return Query(service, entityName.ToString(), attributeName, attributeValue);
        }

        public static FilterExpression BuildCriteria(string attributeName, object attributeValue)
        {
            // Create the ConditionExpression.
            ConditionExpression condition = new ConditionExpression();
            // Set the condition for the retrieval to be 
            // when the account owner's last name is not Cannon.
            condition.AttributeName = attributeName;
            condition.Operator = ConditionOperator.Equal;
            condition.Values = new object[] { attributeValue };

            // Build the filter based on the condition.
            FilterExpression filter = new FilterExpression();
            filter.FilterOperator = LogicalOperator.And;
            filter.Conditions = new ConditionExpression[] { condition };

            return filter;
        }
        
        public static QueryExpression BuildQuery(string entityName, string attributeName, object attributeValue)
        {
            QueryExpression query = BuildQuery(entityName);

            query.Criteria = BuildCriteria(attributeName, attributeValue);

            return query;
        }

        public static BusinessEntityCollection Query(CrmService service, string entityName, string linkToTable)
        {
            return Query(service, BuildQuery(entityName, linkToTable));
        }

        private static LinkEntity BuildLinkEntity(string linkFrom, string linkTo)
        {
            // Create a LinkEntity to link the owner's information to the account.
            LinkEntity link = new LinkEntity();

            // Set the linking entity to be the account.
            link.LinkFromEntityName = linkFrom;

            // Set the entity being linked to to be the system user.
            link.LinkToEntityName = linkTo;

            string fromAttribute = linkFrom + "id";
            string toAttribute = linkFrom + "id";

            // Set the linking attribute to be the owning user.
            link.LinkFromAttributeName = fromAttribute;
            
            // Set the attribute being linked to to be the system user ID.
            link.LinkToAttributeName = toAttribute;

            return link;
        }

        private static QueryExpression BuildQuery(string entityName, string linkToTable)
        {
            QueryExpression query = BuildQuery(entityName);

            LinkEntity link = BuildLinkEntity(entityName, linkToTable);

            query.LinkEntities = new LinkEntity[] { link };

            return query;
        }

        public static BusinessEntityCollection Query(CrmService service, EntityName entityName, EntityName linkToTable, string linkAttributeName, object linkAttributeValue)
        {
            return Query(service, entityName.ToString(), linkToTable.ToString(), linkAttributeName, linkAttributeValue);
        }

        public static BusinessEntityCollection Query(CrmService service, string entityName, string linkToTable, string linkAttributeName, object linkAttributeValue)
        {
            return Query(service, BuildQuery(entityName, linkToTable, linkAttributeName, linkAttributeValue));
        }

        public static QueryExpression BuildQuery(string entityName, string linkToTable, string linkAttributeName, object linkAttributeValue)
        {
            QueryExpression query = BuildQuery(entityName, linkToTable);

            query.LinkEntities[0].LinkCriteria = BuildCriteria(linkAttributeName, linkAttributeValue);

            return query;
        }

        public static DynamicEntity GetDynamicEntity(CrmService service, string entityName, Guid entityId)
        {
            TargetRetrieveDynamic target = new TargetRetrieveDynamic();
            target.EntityId = entityId;
            target.EntityName = entityName;
            RetrieveRequest retrieve = new RetrieveRequest();
            retrieve.Target = target;
            retrieve.ColumnSet = new AllColumns();
            retrieve.ReturnDynamicEntities = true;

            RetrieveResponse response = (RetrieveResponse)service.Execute(retrieve);
            DynamicEntity retrievedEntity = (DynamicEntity)response.BusinessEntity;
            return retrievedEntity;
        }

    }
}
