using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using MsCrmHelpers;
using System.Data;
using System.Data.SqlClient;
using MsCrmHelpers.CrmSdk;
using MsCrmHelpers.CrmMetadataService;
using System.Web.Services.Protocols;

namespace FastSqlSync
{
    public class CrmSynchronizer
    {
        public CrmSynchronizer(CrmServer crmServer, TableReflectingView syncSpec, string crmEntityName)
        {
            m_CrmServer = crmServer;
            m_SyncSpec = syncSpec;
            m_CrmEntityName = crmEntityName;
        }

        private CrmServer m_CrmServer;
        public CrmServer CrmServer
        {
            get { return m_CrmServer; }
        }


        private TableReflectingView m_SyncSpec;
        public TableReflectingView SyncSpec
        {
            get { return m_SyncSpec; }
            set { m_SyncSpec = value; }
        }

        private string m_CrmEntityName;
        public string CrmEntityName
        {
            get { return m_CrmEntityName; }
        }

        private DataSet m_TargetSchemaDataSet;
        public DataSet TargetSchemaDataSet
        {
            get 
            {
                if (m_TargetSchemaDataSet == null)
                {
                    m_TargetSchemaDataSet = new DataSet();
                    SyncSpec.Target.SqlAdapter.FillSchema(m_TargetSchemaDataSet, SchemaType.Source);
                }
                return m_TargetSchemaDataSet;
            }
        }
	
        
        public void DoPreReplicationSync()
        {
            InsertNewRowsFromCrm();

            UpdateChangedRowsFromCrm();

            DeleteRowsRemovedInCrm();
        }

        public void DoPostReplicationSync()
        {
            // Get the TargetSchemaDataSet because can only do one concurrent SQL operation per connection
            DataSet targetSchemaDataSet = TargetSchemaDataSet;

            ProcessRowsInsertedOnWeb();

            ProcessRowsUpdatedOnWeb();
        }


        private void ProcessRowsUpdatedOnWeb()
        {
            DataSet updatedDataSet = new DataSet();

            using (SqlDataAdapter adapter = new SqlDataAdapter(SyncSpec.SelectRowsUpdatedOnWeb, SyncSpec.Target.Connection))
            {
                adapter.Fill(updatedDataSet);
            }

            foreach (DataRow row in updatedDataSet.Tables[0].Rows)
            {
                ProcessRowUpdatedOnWeb(row);
            }
        }

        private void ProcessRowsInsertedOnWeb()
        {
            DataSet updatedDataSet = new DataSet();

            using (SqlDataAdapter adapter = new SqlDataAdapter(SyncSpec.SelectRowsInsertedOnWeb, SyncSpec.Target.Connection))
            {
                adapter.Fill(updatedDataSet);
            }

            foreach (DataRow row in updatedDataSet.Tables[0].Rows)
            {
                ProcessRowInsertedOnWeb(row);
            } 
        }


        private void ProcessRowUpdatedOnWeb(DataRow sqlRow)
        {
            Guid objectId = (Guid)sqlRow[SyncSpec.Source.KeyColumn];

            DynamicEntity crmObject = GetCrmObject(objectId);

            if (crmObject == null)
            {
                ProgressLog.Current.WriteWarning("Detected concurrent delete.  Ignoring updated row on website - should be deleted next cycle");
                return;
            }

            CrmDateTimeProperty crmLastModifiedProperty = (CrmDateTimeProperty) crmObject.GetCrmProperty(SyncSpec.Source.VersionColumn);
            DateTime sqlLastModified = (DateTime)sqlRow[SyncSpec.Target.VersionColumn];
            DateTime crmLastModified = DateTime.Parse(crmLastModifiedProperty.Value.Value).ToUniversalTime();

            if (crmLastModified > sqlLastModified)
            {
                ProgressLog.Current.WriteWarning("Detected concurrent modification.  Ignoring updated row on website - should be deleted next cycle");
                return;
            }
            else if (crmLastModified < sqlLastModified)
            {
                ProgressLog.Current.WriteError("Detected unexpected value for ModifiedOn.  Ignoring updated row on website - may need manual repair");
                return;
            }

            FillCrmObject(sqlRow, crmObject, false);

            CrmServer.CrmService.Update(crmObject);

            CopyBackSingleObject(objectId);
        }

        private void FillCrmObject(DataRow sqlRow, DynamicEntity crmObject, bool isNew)
        {
            EntityMetadata metadata = CrmEntityMetadata;

            CrmPropertyList properties = new CrmPropertyList(crmObject.Properties);
            foreach (AttributeMetadata crmAttribute in metadata.Attributes)
            {
                // Skip attributes we cannot set
                if (isNew)
                {
                    if (!crmAttribute.ValidForCreate) continue;
                }
                else
                {
                    if (!crmAttribute.ValidForUpdate) continue;
                }

                // Ignore nested attributes
                if (!String.IsNullOrEmpty(crmAttribute.AttributeOf))
                {
                    continue;
                }

                string crmAttributeName = crmAttribute.Name;
                switch (crmAttribute.Type)
                {
                    case AttributeType.Virtual:
                        ProgressLog.Current.WriteInfo("Skipped virtual attribute: " + crmAttributeName);
                        break;

                    case AttributeType.String:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                StringProperty property = new StringProperty();
                                property.Name = crmAttributeName;
                                property.Value = (string)sqlValue;
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.UniqueIdentifier:
                    case AttributeType.PrimaryKey:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                UniqueIdentifierProperty property = new UniqueIdentifierProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateUniqueIdentifier((Guid)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Integer:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmNumberProperty property = new CrmNumberProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmNumber((int)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Boolean:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmBooleanProperty property = new CrmBooleanProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmBoolean((bool)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Float:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmFloatProperty property = new CrmFloatProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmFloat((float)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Money:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmMoneyProperty property = new CrmMoneyProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmMoney((decimal)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Decimal:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmDecimalProperty property = new CrmDecimalProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmDecimal((decimal)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.DateTime:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CrmDateTimeProperty property = new CrmDateTimeProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateCrmDateTimeFromUniversal((DateTime)sqlValue);
                                properties.SetProperty(property);
                            }
                            break;
                        }




                    case AttributeType.Lookup:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                LookupProperty property = new LookupProperty();
                                property.Name = crmAttributeName;

                                property.Value = new Lookup();
                                property.Value.Value = (Guid)sqlValue;
                                if (ContainsColumn(sqlRow, crmAttributeName + "Dsc"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Dsc"))
                                        property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                                }
                                if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Name"))
                                        property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                                }
                                if (ContainsColumn(sqlRow, crmAttributeName + "Type"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Type"))
                                        property.Value.type = CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);
                                }

                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Picklist:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                PicklistProperty property = new PicklistProperty();
                                property.Name = crmAttributeName;
                                property.Value = new Picklist();

                                property.Value.Value = (int)sqlValue;
                                if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                                    property.Value.name = (string)sqlRow[crmAttributeName + "Name"];

                                properties.SetProperty(property);

                                //throw new NotImplementedException();

                                //property.Value = (string)sqlValue;
                                //properties.Add(property);
                            }
                            break;
                        }

                    case AttributeType.Owner:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                OwnerProperty property = new OwnerProperty();
                                property.Name = crmAttributeName;
                                property.Value = new Owner();
                                property.Value.Value = (Guid)sqlValue;
                                property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                                property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                                property.Value.type = CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);

                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Status:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                StatusProperty property = new StatusProperty();
                                property.Name = crmAttributeName;
                                property.Value = new Status();
                                property.Value.Value = (int)sqlValue;

                                properties.SetProperty(property);
                            }
                            break;
                        }

                    case AttributeType.Customer:
                        {
                            object sqlValue = sqlRow[crmAttributeName];
                            if ((sqlValue != null) && (sqlValue != DBNull.Value))
                            {
                                CustomerProperty property = new CustomerProperty();
                                property.Name = crmAttributeName;
                                property.Value = new Customer();

                                property.Value.Value = (Guid)sqlValue;
                                if (ContainsColumn(sqlRow, crmAttributeName + "Dsc"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Dsc"))
                                        property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                                }
                                if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Name"))
                                        property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                                }
                                if (ContainsColumn(sqlRow, crmAttributeName + "Type"))
                                {
                                    if (!sqlRow.IsNull(crmAttributeName + "Type"))
                                        property.Value.type = CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);
                                }
                                properties.SetProperty(property);

                                //property.Value = (string)sqlValue;
                                //properties.Add(property);
                            }
                            break;
                        }

                    default:
                        throw new ArgumentOutOfRangeException("AttributeMetadata.Type", crmAttribute.Type, "Encountered unknown AttributeMetadata type");
                }
            }

            crmObject.Properties = properties.ToArray();
        }

        public static SqlParameter CreateInputParam(SqlCommand command,
                                               string paramName,
                                              SqlDbType dbType,
                                              object objValue)
        {
            SqlParameter parameter = CreateInputParam(paramName, dbType, objValue);
            command.Parameters.Add(parameter);
            return parameter;
        }

        public static SqlParameter CreateInputParam(string paramName,
                                               SqlDbType dbType,
                                               object objValue)
        {
            SqlParameter param = new SqlParameter(paramName, dbType);

            if (objValue == null)
            {
                param.IsNullable = true;
                param.Value = DBNull.Value;
            }
            else
            {
                param.Value = objValue;
            }

            return param;
        }

        /// <summary>
        /// Updates a single row in the target table directly from the source
        /// </summary>
        /// <param name="crmId"></param>
        private void CopyBackSingleObject(Guid crmId)
        {
            using (SqlConnection concurrentConnection = new SqlConnection(SyncSpec.Target.Connection.ConnectionString))
            {
                concurrentConnection.Open();
                using (SqlCommand command = concurrentConnection.CreateCommand())
                {
                    command.CommandText = SyncSpec.UpdateSingleRowFromCrm;
                    CreateInputParam(command, "@crmId", SqlDbType.UniqueIdentifier, crmId);
                    int rowsAffected = command.ExecuteNonQuery();
                    if (rowsAffected != 1)
                        throw new ApplicationException("Expected 1 row to be affected by update, but actually was: " + rowsAffected);
                }
            }
        }

        private DynamicEntity GetCrmObject(Guid objectId)
        {
            return QuickQueries.GetDynamicEntity(CrmServer.CrmService, CrmEntityName, objectId);
        }

        private void ProcessRowInsertedOnWeb(DataRow sqlRow)
        {
            DynamicEntity crmNewObject = new DynamicEntity();
            crmNewObject.Name = CrmEntityName;

            FillCrmObject(sqlRow, crmNewObject, true);

            ProgressLog.Current.WriteInfo("Creating object in CRM");
            Guid newCrmId;
            try
            {
                newCrmId = CrmServer.CrmService.Create(crmNewObject);
            }
            catch (SoapException soapException)
            {
                ProgressLog.Current.WriteError("Failed to create object in CRM: " + crmNewObject.Name);
                foreach (Property property in crmNewObject.Properties)
                {
                    ProgressLog.Current.WriteInfo("\t" + property.ToString());
                }
                ProgressLog.Current.WriteError(soapException.Detail.OuterXml);
                return;
            }

            CopyBackSingleObject(newCrmId);
        }


        private bool ContainsColumn(SqlDataReader sqlReader, string name)
        {
            for (int i = 0; i < sqlReader.FieldCount; i++)
            {
                if (sqlReader.GetName(i).Equals(name, StringComparison.InvariantCultureIgnoreCase))
                    return true;
            }

            return false;
        }

        private bool ContainsColumn(DataRow sqlRow, string name)
        {
            return sqlRow.Table.Columns.Contains(name);
        }

        private void DeleteRowsRemovedInCrm()
        {
            int rowCount = ExecuteNonQuery(SyncSpec.DeleteRowsRemovedInCrm);
            ProgressLog.Current.WriteInfo("Rows deleted: " + rowCount);
        }

        private void UpdateChangedRowsFromCrm()
        {
            int rowCount = ExecuteNonQuery(SyncSpec.UpdateChangedRowsFromCrm);
            ProgressLog.Current.WriteInfo("Rows updated: " + rowCount);
        }

        private void InsertNewRowsFromCrm()
        {
            int rowCount = ExecuteNonQuery(SyncSpec.InsertNewRowsFromCrm);
            ProgressLog.Current.WriteInfo("Rows inserted: " + rowCount);
        }


        private int ExecuteNonQuery(string sql)
        {
            using (SqlCommand command = SyncSpec.Target.Connection.CreateCommand())
            {
                command.CommandText = sql;
                return command.ExecuteNonQuery();
            }
        }

        EntityMetadata m_CrmEntityMetadata;
        public EntityMetadata CrmEntityMetadata
        {
            get
            {
                if (m_CrmEntityMetadata == null)
                    m_CrmEntityMetadata = CrmServer.RetrieveEntityMetadata(CrmEntityName);
                return m_CrmEntityMetadata;
            }
        }

        public Metadata CrmMetadata
        {
            get
            {
               return CrmServer.CrmMetadata;
            }
        }
    }
}
