using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using MsCrmHelpers.CrmMetadataService;
using MsCrmHelpers;
using System.Diagnostics;
using MsCrmHelpers.CrmSdk;
using System.Web.Services.Protocols;
using System.Text.RegularExpressions;

namespace FastSqlSync
{
    /// <summary>
    /// A mapping between a CRM field and a SQL field
    /// </summary>
    public class FieldMap
    {
        private string m_CrmField;

        public string CrmField
        {
            get { return m_CrmField; }
            set { m_CrmField = value; }
        }

        private string m_SqlField;
        public string SqlField
        {
            get { return m_SqlField; }
            set { m_SqlField = value; }
        }

        private string m_StringFormat = "{0}";

        public string StringFormat
        {
            get { return m_StringFormat; }
            set { m_StringFormat = value; }
        }

        private string m_Extract;
        public string Extract
        {
            get { return m_Extract; }
            set { m_Extract = value; RegexPattern = BuildRegexFromExtract(value); }
        }

        private string m_RegexPattern;
        public string RegexPattern
        {
            get
            {
                return m_RegexPattern;
            }
            set { m_RegexPattern = value; }
        }

        //private bool m_AutoTruncate;

        ///// <summary>
        ///// Set to automatically truncate strings to the column length
        ///// </summary>
        //public bool AutoTruncate
        //{
        //    get { return m_AutoTruncate; }
        //    set { m_AutoTruncate = value; }
        //}
	
        protected internal Regex RegexObject
        {
            get
            {
                string regex = RegexPattern;
                if (String.IsNullOrEmpty(regex))
                    return null;
                return new Regex(regex, RegexOptions.IgnoreCase);
            }
        }

        public FieldMap()
        {
        }

        /// <summary>
        /// Creates a regex from an extract string
        /// </summary>
        /// <param name="extract"></param>
        /// <returns></returns>
        private static string BuildRegexFromExtract(string extract)
        {
            extract = Regex.Escape(extract);

            // The star is our wildcard
            extract = extract.Replace("\\*", "(.*)");

            return extract;
        }

        public const string SpecialColumnExistsInCrm = "xfer__ExistsInCrm";
        public const string SpecialColumnChangedByWeb = "xfer__ChangedByWeb";


        internal void CopySqlToCrm(DataRow sqlRow, EntityMetadata entityMetadata, CrmPropertyList properties, bool alreadyExists, EntityCache entityCache)
        {
            AttributeMetadata crmAttribute = entityMetadata.FindAttributeMetadata(CrmField);
            if (crmAttribute == null) throw new ArgumentOutOfRangeException("CrmField", CrmField, "Field not found");

            // Skip attributes we cannot set
            if (alreadyExists)
            {
                if (!crmAttribute.ValidForUpdate) return;
            }
            else
            {
                if (!crmAttribute.ValidForCreate) return;
            }

            // Ignore nested attributes
            if (!String.IsNullOrEmpty(crmAttribute.AttributeOf))
            {
                return;
            }

            string crmAttributeName = crmAttribute.Name;
            switch (crmAttribute.Type)
            {
                case AttributeType.Virtual:
                    ProgressLog.Current.WriteInfo("Skipped virtual attribute: " + crmAttributeName);
                    break;

                case AttributeType.String:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            string stringValue = null;
                            if (!String.IsNullOrEmpty(RegexPattern))
                            {
                                Regex regexObject = RegexObject;
                                Match regexMatch = regexObject.Match((string)sqlValue);
                                if (regexMatch.Success)
                                {
                                    stringValue = regexMatch.Groups[1].Value;
                                    stringValue = String.Format(StringFormat, stringValue);
                                }
                            }
                            else
                            {
                                stringValue = String.Format(StringFormat, sqlValue);
                            }

                            //if (AutoTruncate)
                            //{
                            //    if (!String.IsNullOrEmpty(stringValue) && (stringValue.Length > ((StringAttributeMetadata) crmAttribute).MaxLength))
                            //    {
                            //        stringValue = stringValue.Substring(0, ((StringAttributeMetadata)crmAttribute).MaxLength);
                            //    }
                            //}

                            StringProperty property = new StringProperty();
                            property.Name = crmAttributeName;
                            property.Value = stringValue;
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.UniqueIdentifier:
                case AttributeType.PrimaryKey:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            Guid sqlGuid = (Guid)sqlValue;
                            if (sqlGuid != Guid.Empty)
                            {
                                UniqueIdentifierProperty property = new UniqueIdentifierProperty();
                                property.Name = crmAttributeName;
                                property.Value = CrmTypes.CreateUniqueIdentifier(sqlGuid);
                                properties.SetProperty(property);
                            }
                        }
                        break;
                    }

                case AttributeType.Integer:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmNumberProperty property = new CrmNumberProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmNumber((int)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Boolean:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmBooleanProperty property = new CrmBooleanProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmBoolean((bool)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Float:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmFloatProperty property = new CrmFloatProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmFloat((float)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Money:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmMoneyProperty property = new CrmMoneyProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmMoney((decimal)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Decimal:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmDecimalProperty property = new CrmDecimalProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmDecimal((decimal)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.DateTime:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CrmDateTimeProperty property = new CrmDateTimeProperty();
                            property.Name = crmAttributeName;
                            property.Value = CrmTypes.CreateCrmDateTimeFromUniversal((DateTime)sqlValue);
                            properties.SetProperty(property);
                        }
                        break;
                    }




                case AttributeType.Lookup:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            Guid sqlGuid = (Guid)sqlValue;
                            if (sqlGuid != Guid.Empty)
                            {
                                LookupProperty property = new LookupProperty();
                                property.Name = crmAttributeName;

                                property.Value = new Lookup();
                                property.Value.Value = sqlGuid;

                                RelationshipMetadata reference = null;
                                foreach (RelationshipMetadata referenceFrom in entityMetadata.ReferencesFrom)
                                {
                                    if (referenceFrom.ReferencingAttribute == crmAttribute.Name)
                                    {
                                        reference = referenceFrom;
                                        break;
                                    }
                                }
                                if (reference == null) throw new ArgumentOutOfRangeException("CrmAttribute", crmAttribute.Name, "Reference not found for Lookup");

                                DynamicEntity referencedEntity = referencedEntity = entityCache.GetEntity(reference.ReferencedEntity, property.Value.Value);

                                if (referencedEntity == null) throw new ArgumentOutOfRangeException(SqlField, sqlValue, "Referenced object not found in CRM");

                                property.Value.dsc = 0;
                                property.Value.name = referencedEntity.Name;
                                property.Value.type = reference.ReferencedEntity;

                                ////if (AttributeExists(crmAttribute, crmAttributeName + "Dsc"))
                                //{
                                //    property.Value.dsc = 0;

                                //    //if (ContainsColumn(sqlRow, crmAttributeName + "Dsc"))
                                //    //{
                                //    //    if (!sqlRow.IsNull(crmAttributeName + "Dsc"))
                                //    //        property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                                //    //}
                                //}

                                //if (AttributeExists(crmAttribute, crmAttributeName + "Name"))
                                //{

                                //    //if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                                //    //{
                                //    //    if (!sqlRow.IsNull(crmAttributeName + "Name"))
                                //    //        property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                                //    //}
                                //    throw new NotImplementedException();
                                //}

                                //if (AttributeExists(crmAttribute, crmAttributeName + "Type"))
                                //{
                                //    //if (ContainsColumn(sqlRow, crmAttributeName + "Type"))
                                //    //{
                                //    //    if (!sqlRow.IsNull(crmAttributeName + "Type"))
                                //    //        property.Value.type = CrmServer.Current.CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);
                                //    //}
                                //    throw new NotImplementedException();
                                //}
                                properties.SetProperty(property);
                            }
                        }
                        break;
                    }

                case AttributeType.Picklist:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            PicklistProperty property = new PicklistProperty();
                            property.Name = crmAttributeName;
                            property.Value = new Picklist();

                            property.Value.Value = (int)sqlValue;
                            if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                                property.Value.name = (string)sqlRow[crmAttributeName + "Name"];

                            properties.SetProperty(property);

                            //throw new NotImplementedException();

                            //property.Value = (string)sqlValue;
                            //properties.Add(property);
                        }
                        break;
                    }

                case AttributeType.Owner:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            OwnerProperty property = new OwnerProperty();
                            property.Name = crmAttributeName;
                            property.Value = new Owner();
                            property.Value.Value = (Guid)sqlValue;
                            property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                            property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                            property.Value.type = CrmServer.Current.CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);

                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Status:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            StatusProperty property = new StatusProperty();
                            property.Name = crmAttributeName;
                            property.Value = new Status();
                            property.Value.Value = (int)sqlValue;

                            properties.SetProperty(property);
                        }
                        break;
                    }

                case AttributeType.Customer:
                    {
                        object sqlValue = sqlRow[SqlField];
                        if ((sqlValue != null) && (sqlValue != DBNull.Value))
                        {
                            CustomerProperty property = new CustomerProperty();
                            property.Name = crmAttributeName;
                            property.Value = new Customer();

                            property.Value.Value = (Guid)sqlValue;
                            if (ContainsColumn(sqlRow, crmAttributeName + "Dsc"))
                            {
                                if (!sqlRow.IsNull(crmAttributeName + "Dsc"))
                                    property.Value.dsc = (int)sqlRow[crmAttributeName + "Dsc"];
                            }
                            if (ContainsColumn(sqlRow, crmAttributeName + "Name"))
                            {
                                if (!sqlRow.IsNull(crmAttributeName + "Name"))
                                    property.Value.name = (string)sqlRow[crmAttributeName + "Name"];
                            }
                            if (ContainsColumn(sqlRow, crmAttributeName + "Type"))
                            {
                                if (!sqlRow.IsNull(crmAttributeName + "Type"))
                                    property.Value.type = CrmServer.Current.CrmMetadata.MapTypeCodeToEntityName((int)sqlRow[crmAttributeName + "Type"]);
                            }
                            properties.SetProperty(property);

                            //property.Value = (string)sqlValue;
                            //properties.Add(property);
                        }
                        break;
                    }

                default:
                    throw new ArgumentOutOfRangeException("AttributeMetadata.Type", crmAttribute.Type, "Encountered unknown AttributeMetadata type");
            }
        }


        private static bool ContainsColumn(DataRow sqlRow, string name)
        {
            return sqlRow.Table.Columns.Contains(name);
        }

        /// <summary>
        /// Builds a SQL query that compares the CRM view to the SQL table, to detect changed rows
        /// </summary>
        /// <param name="crmTableName"></param>
        /// <param name="sqlTableName"></param>
        /// <returns></returns>
        internal string ConstructSqlDifferenceClause(EntityMetadata entityMetadata, string crmTableName, string sqlTableName)
        {
            AttributeMetadata attributeMetadata = entityMetadata.FindAttributeMetadata(CrmField);
            if (attributeMetadata == null) throw new ArgumentOutOfRangeException("CrmField", CrmField, "Field not found");

            string crmFQ = crmTableName + "." + CrmField;
            string sqlFQ = sqlTableName + "." + SqlField;

            string defaultValue = "''";
            switch (attributeMetadata.Type)
            {
                    // Though not a GUID, as a dummy value this works well
                case AttributeType.String:

                case AttributeType.Lookup:
                case AttributeType.PrimaryKey:
                case AttributeType.UniqueIdentifier:
                    {
                        defaultValue = "'" + Guid.NewGuid().ToString() + "'";
                    }
                    break;

                case AttributeType.DateTime:
                    {
                        defaultValue = "'" + DateTime.Now.ToString() + "'";
                    }
                    break;

                default:
                    //Debug.WriteLine(attributeMetadata.Type);
                    break;
            }

            
            string clause = "ISNULL(" + crmFQ + ", " + defaultValue + ") <> ISNULL(" + sqlFQ + ", " + defaultValue + ")";
            
            switch (attributeMetadata.Type)
            {
                case AttributeType.String:
                    {
                        StringAttributeMetadata stringMetadata = (StringAttributeMetadata)attributeMetadata;
                        if (stringMetadata.MaxLength > 1999)
                        {
                            // Assume ntext
                            ProgressLog.Current.WriteInfo("Ignoring long string column: " + CrmField + " vs " + SqlField);
                            clause = "0 = 1";
                        }
                        else
                        {
                            clause += " COLLATE Latin1_General_BIN ";
                        }
                    }
                    break;

                case AttributeType.DateTime:
                    clause = "DATEDIFF(s, ISNULL(" + crmFQ + ", " + defaultValue + "), ISNULL(" + sqlFQ + ", " + defaultValue + ")) <> 0";
                    break;
               
            }

            if (StringFormat != "{0}")
            {
                ProgressLog.Current.WriteInfo("Not comparing formatted column: " + CrmField + " vs " + SqlField);
                clause = "0 = 1";
            }

            if (!String.IsNullOrEmpty(RegexPattern))
            {
                ProgressLog.Current.WriteInfo("Not comparing regex column: " + CrmField + " vs " + SqlField);
                clause = "0 = 1";
            }

            return "(" + clause + ")";
        }

    }
}
