using System;
using System.Collections.Generic;
using System.Text;
using System.Data.SqlClient;
using System.Data;
using Microsoft.SqlServer.Management.Smo;
using Microsoft.SqlServer.Management.Common;

namespace FastSqlSync
{
    /// <summary>
    /// A connection to a table, with information about the key and version columns 
    /// </summary>
    public class SyncTableSpecification
    {
        public SyncTableSpecification(SqlConnection connection, string tableName, string keyColumn, string versionColumn)
        {
            if (connection == null)
                throw new ArgumentNullException();
            if (String.IsNullOrEmpty(tableName))
                throw new ArgumentNullException();
            if (String.IsNullOrEmpty(keyColumn))
                throw new ArgumentNullException();

            m_Connection = connection;
            m_TableName = tableName;
            m_KeyColumn = keyColumn;
            m_VersionColumn = versionColumn;
        }

        private void BuildCommands()
        {
            SqlCommand selectAllCommand = CreateCommand("SELECT TOP 1 * FROM " + m_TableName);
            DataTable schemaTable = new DataTable();
            using (SqlDataAdapter adapter = new SqlDataAdapter(selectAllCommand))
            {
                adapter.FillSchema(schemaTable, SchemaType.Source);
            }
            StringBuilder updateFragmentBuilder = new StringBuilder();
            StringBuilder columnListBuilder = new StringBuilder();
            StringBuilder parameterListBuilder = new StringBuilder();
            foreach (DataColumn column in schemaTable.Columns)
            {
                if (updateFragmentBuilder.Length != 0)
                    updateFragmentBuilder.Append(", ");
                if (columnListBuilder.Length != 0)
                    columnListBuilder.Append(", ");
                if (parameterListBuilder.Length != 0)
                    parameterListBuilder.Append(", ");

                updateFragmentBuilder.AppendFormat("[{0}]=@p_{0}", column.ColumnName);

                columnListBuilder.Append("[" + column.ColumnName + "]");
                parameterListBuilder.Append("@p_" + column.ColumnName);
            }

            string updateFragment = updateFragmentBuilder.ToString();
            string columnList = columnListBuilder.ToString();
            string parameterList = parameterListBuilder.ToString();
            string whereSql = "WHERE [" + m_KeyColumn + "]=" + "@p_" + m_KeyColumn;

            string updateSql = String.Format("UPDATE {0} SET {1} {2}", m_TableName, updateFragment, whereSql);
            string insertSql = String.Format("INSERT {0} ({1}) VALUES ({2})", m_TableName, columnList, parameterList);
            string deleteSql = String.Format("DELETE FROM {0} {1}", m_TableName, whereSql);

            string updateOrInsertSql = updateSql + "; IF @@ROWCOUNT=0 " + insertSql;
            m_InsertUpdateCommand = CreateCommand(updateOrInsertSql);

            m_DeleteCommand = CreateCommand(deleteSql);
            m_DeleteCommand.Parameters.Add("@p_" + m_KeyColumn, GetSqlDbType(schemaTable.Columns[m_KeyColumn].DataType));

            m_InsertUpdateCommand = CreateCommand(updateOrInsertSql);
            foreach (DataColumn column in schemaTable.Columns)
            {
                m_InsertUpdateCommand.Parameters.Add("@p_" + column.ColumnName, GetSqlDbType(column.DataType));
            }
        }

        SqlCommand m_InsertUpdateCommand;
        protected SqlCommand InsertUpdateCommand
        {
            get
            {
                if (m_InsertUpdateCommand == null)
                {
                    BuildCommands();
                }
                return m_InsertUpdateCommand;
            }
        }

        SqlCommand m_DeleteCommand;
        protected SqlCommand DeleteCommand
        {
            get
            {
                if (m_DeleteCommand == null)
                {
                    BuildCommands();
                }
                return m_DeleteCommand;
            }
        }

        private SqlConnection m_Connection;
        public SqlConnection Connection
        {
            get { return m_Connection; }
        }

        private string m_TableName;
        public string TableName
        {
            get { return m_TableName; }
        }


        private string m_KeyColumn;
        public string KeyColumn
        {
            get { return m_KeyColumn; }
        }

        private string m_VersionColumn;
        public string VersionColumn
        {
            get { return m_VersionColumn; }
        }

        public SqlCommand CreateCommand(string sql)
        {
            if (sql == null)
                throw new ArgumentNullException();

            if (Connection.State != ConnectionState.Open)
                Connection.Open();

            SqlCommand command = Connection.CreateCommand();
            command.CommandText = sql;
            command.CommandType = CommandType.Text;
            return command;
        }

        protected SqlCommand CreateCommand(string sqlFormat, params object[] args)
        {
            string sql = String.Format(sqlFormat, args);
            return CreateCommand(sql);
        }

        public object GetMaxVersion()
        {
            if (VersionColumn == null)
                throw new ArgumentNullException();

            using (SqlCommand getMaxVersionCommand = CreateCommand("SELECT MAX([{0}]) FROM {1}", VersionColumn, TableName))
            {
                object maxVersion = getMaxVersionCommand.ExecuteScalar();
                return maxVersion;
            }
        }

        public int GetRowCount()
        {
            using (SqlCommand sqlCommand = CreateCommand("SELECT COUNT(*) FROM {0}", TableName))
            {
                int rowCount = (int)sqlCommand.ExecuteScalar();
                return rowCount;
            }
        }

        private SqlDataAdapter m_SqlAdapter;
        public SqlDataAdapter SqlAdapter
        {
            get
            {
                if (m_SqlAdapter == null)
                {
                    m_SqlAdapter = new SqlDataAdapter("SELECT * FROM " + TableName, Connection);
                    SqlCommandBuilder commandBuilder = new SqlCommandBuilder(m_SqlAdapter);
                }

                return m_SqlAdapter; 
            }
        }
	
        public static SqlDbType GetSqlDbType(Type type)
        {
            if (type == null)
                throw new ArgumentNullException();

            if (type == typeof(Guid))
            {
                return SqlDbType.UniqueIdentifier;
            }

            if (type == typeof(DateTime))
            {
                return SqlDbType.DateTime;
            }

            if (type == typeof(string))
            {
                return SqlDbType.NVarChar;
            }

            if (type == typeof(Int32))
            {
                return SqlDbType.Int;
            }

            if (type == typeof(Boolean))
            {
                return SqlDbType.Bit;
            }

            if (type == typeof(Double))
            {
                return SqlDbType.Float;
            }

            if (type == typeof(Decimal))
            {
                return SqlDbType.Decimal;
            }

            if (type == typeof(Byte[]))
            {
                return SqlDbType.Image;
            }

            if (type == typeof(Int64))
            {
                return SqlDbType.BigInt;
            }

            throw new ArgumentOutOfRangeException();
        }

        public SqlDataReader GetNewerRows(object minVersion)
        {
            if (VersionColumn == null)
                throw new ArgumentNullException();

            if ((minVersion == null) || (minVersion == DBNull.Value))
                return GetAllRows();

            using (SqlCommand sqlCommand = CreateCommand("SELECT * FROM {0} WHERE [{1}] >= @minVersion ORDER BY {1} ASC", TableName, VersionColumn))
            {
                sqlCommand.Parameters.Add("@minVersion", GetSqlDbType(minVersion.GetType()));
                sqlCommand.Parameters[0].Value = minVersion;

                return sqlCommand.ExecuteReader();
            }
        }


        public SqlDataReader GetAllRows()
        {
            using (SqlCommand sqlCommand = CreateCommand("SELECT * FROM {0}", TableName))
            {
                return sqlCommand.ExecuteReader();
            }
        }

        public void ProcessRowChange(SqlDataReader sourceChangeRow)
        {
            SqlCommand insertUpdateCommand = InsertUpdateCommand;
            for (int i = 0; i < sourceChangeRow.FieldCount; i++)
            {
                SqlParameter parameter = insertUpdateCommand.Parameters["@p_" + sourceChangeRow.GetName(i)];
                parameter.Value = sourceChangeRow[i];
            }

            int insertUpdatedRows = insertUpdateCommand.ExecuteNonQuery();
            if (insertUpdatedRows == 1)
                return;

            throw new ApplicationException();
        }

        public IEnumerable<object> GetAllKeysSorted()
        {
            List<object> keys = new List<object>();
            SqlCommand sqlCommand = CreateCommand("SELECT {0} FROM {1}", KeyColumn, TableName);
            using (SqlDataReader reader = sqlCommand.ExecuteReader())
            {
                while (reader.Read())
                {
                    keys.Add(reader[0]);
                }
            }
            keys.Sort();

            return keys;
        }

        public void DeleteRow(SqlTransaction transaction, object targetKey)
        {
            SqlCommand deleteCommand = DeleteCommand;
            deleteCommand.Transaction = transaction;

            SqlParameter parameter = deleteCommand.Parameters["@p_" + KeyColumn];
            parameter.Value = targetKey;

            int rowsDeleted = deleteCommand.ExecuteNonQuery();
            if (rowsDeleted == 1)
                return;

            throw new ApplicationException();
        }

        Database m_SmoDatabase;
        internal Database SmoDatabase
        {
            get
            {
                if (m_SmoDatabase == null)
                {
                    m_SmoDatabase = SmoServer.Databases[Connection.Database];
                }
                return m_SmoDatabase;
            }
        }

        Server m_SmoServer;
        internal Server SmoServer
        {
            get
            {
                if (m_SmoServer == null)
                {
                    m_SmoServer = new Server(new ServerConnection(Connection));
                }
                return m_SmoServer;
            }
        }
    }
}
