using System;
using System.Collections.Generic;
using System.Text;
using System.Data;

using Microsoft.SqlServer.Management.Smo;
using Microsoft.SqlServer.Management.Smo.Agent;
using Microsoft.SqlServer.Management.Common;
using System.Data.SqlClient;
using System.Diagnostics;
using System.Collections.Specialized;


namespace FastSqlSync
{
    /// <summary>
    /// Creates a SQL table to hold the data in the specified schema table
    /// Uses SMO, so depends on SQL 2005
    /// </summary>
    public class SyncSchema
    {
        private SyncTableSpecification m_Source;
        public SyncTableSpecification Source
        {
            get { return m_Source; }
        }

        private SyncTableSpecification m_Target;
        public SyncTableSpecification Target
        {
            get { return m_Target; }
        }

        public SyncSchema(SyncTableSpecification source, SyncTableSpecification target)
        {
            m_Source = source;
            m_Target = target;
        }


        private Table GetSmoTable(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            Table table = database.Tables[connection.TableName];
            return table;
        }

        private View GetSmoView(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            View view = database.Views[connection.TableName];
            return view;
        }

        private TableViewBase GetSmoTableOrView(SyncTableSpecification connection)
        {
            Table table = GetSmoTable(connection);
            if (table != null) return table;

            View view = GetSmoView(connection);
            if (view != null) return view;

            throw new ArgumentOutOfRangeException();
        }

        private Table CreateSmoTable(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            Table table = new Table(database, connection.TableName);
            return table;
        }

        public void CopyTableSchema()
        {
            Source.SmoServer.SetDefaultInitFields(typeof(Column), true);
            Target.SmoServer.SetDefaultInitFields(typeof(Column), true);
            
            TableViewBase sourceTable = GetSmoTableOrView(Source);
            Table targetTable = GetSmoTable(Target);

            if (targetTable == null)
            {
                targetTable = CreateSmoTable(Target);
            }

            ProgressLog.Current.WriteInfo(targetTable.Name);

            if (sourceTable.Columns.Count == targetTable.Columns.Count)
            {
                ProgressLog.Current.WriteInfo("\tSame number of columns - skipped");
                return;
            }

            bool isNewTable = (targetTable.State == SqlSmoState.Creating);

            foreach (Column sourceColumn in sourceTable.Columns)
            {
                Column targetColumn = targetTable.Columns[sourceColumn.Name];
                if (targetColumn == null)
                {
                    targetColumn = new Column(targetTable, sourceColumn.Name);
                    DataType targetColumnType = sourceColumn.DataType;

                    if (targetColumnType.SqlDataType == SqlDataType.Timestamp)
                    {
                        if (sourceColumn.Nullable)
                        {
                            targetColumnType.SqlDataType = SqlDataType.Binary;
                            targetColumnType.MaximumLength = 8;
                        }
                        else
                        {
                            targetColumnType.SqlDataType = SqlDataType.VarBinary;
                            targetColumnType.MaximumLength = 8;
                        }
                    }

                    targetColumn.DataType = targetColumnType;
                }

                ProgressLog.Current.WriteInfo("\t" + sourceColumn.Name);
                //targetColumn.AnsiPaddingStatus = sourceColumn.AnsiPaddingStatus;
                if (isNewTable || (targetColumn.Collation != sourceColumn.Collation))
                    targetColumn.Collation = sourceColumn.Collation;
                if (isNewTable || (targetColumn.Computed != sourceColumn.Computed))
                    targetColumn.Computed = sourceColumn.Computed;
                if (targetColumn.Computed)
                {
                    if (isNewTable || (targetColumn.ComputedText != sourceColumn.ComputedText))
                        targetColumn.ComputedText = sourceColumn.ComputedText;
                   if (isNewTable || (targetColumn.IsPersisted != sourceColumn.IsPersisted))
                        targetColumn.IsPersisted = sourceColumn.IsPersisted;
                }

               if (isNewTable || (targetColumn.Default != sourceColumn.Default))
                    targetColumn.Default = sourceColumn.Default;
                //targetColumn.DefaultConstraint = sourceColumn.DefaultConstraint;
                if (isNewTable || (targetColumn.DefaultSchema != sourceColumn.DefaultSchema))
                    targetColumn.DefaultSchema = sourceColumn.DefaultSchema;
                if (isNewTable || (targetColumn.Identity != sourceColumn.Identity))
                    targetColumn.Identity = sourceColumn.Identity;
                if (isNewTable || (targetColumn.IdentityIncrement != sourceColumn.IdentityIncrement))
                    targetColumn.IdentityIncrement = sourceColumn.IdentityIncrement;
                if (isNewTable || (targetColumn.IdentitySeed != sourceColumn.IdentitySeed))
                    targetColumn.IdentitySeed = sourceColumn.IdentitySeed;
                //targetColumn.IsFullTextIndexed = sourceColumn.IsFullTextIndexed;
                if (isNewTable || (targetColumn.NotForReplication != sourceColumn.NotForReplication))
                    targetColumn.NotForReplication = sourceColumn.NotForReplication;
                if (isNewTable || (targetColumn.Nullable != sourceColumn.Nullable))
                    targetColumn.Nullable = sourceColumn.Nullable;
                if (isNewTable || (targetColumn.RowGuidCol != sourceColumn.RowGuidCol))
                    targetColumn.RowGuidCol = sourceColumn.RowGuidCol;
                if (isNewTable || (targetColumn.Rule != sourceColumn.Rule))
                    targetColumn.Rule = sourceColumn.Rule;
                if (isNewTable || (targetColumn.RuleSchema != sourceColumn.RuleSchema))
                    targetColumn.RuleSchema = sourceColumn.RuleSchema;

                //foreach (Property sourceProperty in sourceColumn.Properties)
                //{
                //    if (!sourceProperty.Writable) continue;

                //    targetColumn.Properties[sourceProperty.Name].Value = sourceProperty.Value;
                //}

                switch (targetColumn.State)
                {
                    case SqlSmoState.Creating:
                        targetTable.Columns.Add(targetColumn);
                        break;
                    case SqlSmoState.Existing:
                        // No action needed
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }

            if (targetTable.State == SqlSmoState.Creating)
            {
                targetTable.Create();
            }
            else
            {
                targetTable.Alter();
            }
        }


        public void CopyStoredProcedures()
        {
            Source.SmoServer.SetDefaultInitFields(typeof(StoredProcedure), true);
            Target.SmoServer.SetDefaultInitFields(typeof(StoredProcedure), true);
            
            Database sourceSmoDatabase = Source.SmoDatabase;
            Database targetSmoDatabase = Target.SmoDatabase;
            ScriptingOptions scriptOptions = new ScriptingOptions();
            foreach (StoredProcedure sourceSproc in sourceSmoDatabase.StoredProcedures)
            {
                ProgressLog.Current.WriteInfo(sourceSproc.Name);
                if (sourceSproc.IsSystemObject)
                {
                    ProgressLog.Current.WriteInfo("\tSkipped");
                    continue;
                }
                StringCollection script = sourceSproc.Script(scriptOptions);
                targetSmoDatabase.ExecuteNonQuery(script);
            }
            //    StoredProcedure targetSproc = targetSmoDatabase.StoredProcedures[sourceSproc.Name];
            //    bool isNewSproc = false;
            //    if (targetSproc == null)
            //    {
            //        targetSproc = new StoredProcedure(targetSmoDatabase, sourceSproc.Name);
            //        isNewSproc = true;
            //    }

            //    Debug.WriteLine("\t" + targetSproc.Name);

            //    targetSproc.
            //    //targetColumn.AnsiPaddingStatus = sourceColumn.AnsiPaddingStatus;
            //    if (isNewTable || (targetColumn.Collation != sourceColumn.Collation))
            //        targetColumn.Collation = sourceColumn.Collation;
            //    if (isNewTable || (targetColumn.Computed != sourceColumn.Computed))
            //        targetColumn.Computed = sourceColumn.Computed;
            //    if (targetColumn.Computed)
            //    {
            //        if (isNewTable || (targetColumn.ComputedText != sourceColumn.ComputedText))
            //            targetColumn.ComputedText = sourceColumn.ComputedText;
            //        if (isNewTable || (targetColumn.IsPersisted != sourceColumn.IsPersisted))
            //            targetColumn.IsPersisted = sourceColumn.IsPersisted;
            //    }

            //    if (isNewTable || (targetColumn.Default != sourceColumn.Default))
            //        targetColumn.Default = sourceColumn.Default;
            //    //targetColumn.DefaultConstraint = sourceColumn.DefaultConstraint;
            //    if (isNewTable || (targetColumn.DefaultSchema != sourceColumn.DefaultSchema))
            //        targetColumn.DefaultSchema = sourceColumn.DefaultSchema;
            //    if (isNewTable || (targetColumn.Identity != sourceColumn.Identity))
            //        targetColumn.Identity = sourceColumn.Identity;
            //    if (isNewTable || (targetColumn.IdentityIncrement != sourceColumn.IdentityIncrement))
            //        targetColumn.IdentityIncrement = sourceColumn.IdentityIncrement;
            //    if (isNewTable || (targetColumn.IdentitySeed != sourceColumn.IdentitySeed))
            //        targetColumn.IdentitySeed = sourceColumn.IdentitySeed;
            //    //targetColumn.IsFullTextIndexed = sourceColumn.IsFullTextIndexed;
            //    if (isNewTable || (targetColumn.NotForReplication != sourceColumn.NotForReplication))
            //        targetColumn.NotForReplication = sourceColumn.NotForReplication;
            //    if (isNewTable || (targetColumn.Nullable != sourceColumn.Nullable))
            //        targetColumn.Nullable = sourceColumn.Nullable;
            //    if (isNewTable || (targetColumn.RowGuidCol != sourceColumn.RowGuidCol))
            //        targetColumn.RowGuidCol = sourceColumn.RowGuidCol;
            //    if (isNewTable || (targetColumn.Rule != sourceColumn.Rule))
            //        targetColumn.Rule = sourceColumn.Rule;
            //    if (isNewTable || (targetColumn.RuleSchema != sourceColumn.RuleSchema))
            //        targetColumn.RuleSchema = sourceColumn.RuleSchema;

            //    //foreach (Property sourceProperty in sourceColumn.Properties)
            //    //{
            //    //    if (!sourceProperty.Writable) continue;

            //    //    targetColumn.Properties[sourceProperty.Name].Value = sourceProperty.Value;
            //    //}

            //    switch (targetColumn.State)
            //    {
            //        case SqlSmoState.Creating:
            //            targetTable.Columns.Add(targetColumn);
            //            break;
            //        case SqlSmoState.Existing:
            //            // No action needed
            //            break;
            //        default:
            //            throw new ArgumentOutOfRangeException();
            //    }
            //}

            //if (targetTable.State == SqlSmoState.Creating)
            //{
            //    targetTable.Create();
            //}
            //else
            //{
            //    targetTable.Alter();
            //}
        }

    }
}
