using System;
using System.Collections.Generic;
using System.Text;
using System.Data.SqlClient;
using System.Xml.Serialization;
using System.IO;
using MsCrmHelpers;

namespace FastSqlSync
{
    /// <summary>
    /// The main control-script that specifies the steps to take as part of a synchronization
    /// </summary>
    public class SyncScript
    {
        private string m_CrmSqlConnection;

        /// <summary>
        /// The SQL connection string to the CRM database
        /// </summary>
        public string CrmSqlConnection
        {
            get { return m_CrmSqlConnection; }
            set { m_CrmSqlConnection = value; }
        }

        private string m_ReplicationPublicationName;

   
        /// <summary>
        /// If this is set, the specified database merge replication will be published to all subscribers
        /// </summary>
        public string ReplicationPublicationName
        {
            get { return m_ReplicationPublicationName; }
            set { m_ReplicationPublicationName = value; }
        }
	
        private string m_TargetSqlConnection;

        /// <summary>
        /// The SQL connection string to the database which holds a copy of the CRM tables
        /// </summary>
        public string TargetSqlConnection
        {
            get { return m_TargetSqlConnection; }
            set { m_TargetSqlConnection = value; }
        }

        private bool m_SyncSchema;

        /// <summary>
        /// Should we sync the schema (i.e. create the tables, add missing fields etc)
        /// </summary>
        public bool SyncSchema
        {
            get { return m_SyncSchema; }
            set { m_SyncSchema = value; }
        }
	
        private List<TableReflectingView> m_SyncTables = new List<TableReflectingView>();

        public List<TableReflectingView> SyncTables
        {
            get { return m_SyncTables; }
            set { m_SyncTables = value; }
        }

        public static readonly XmlSerializer Serializer = new XmlSerializer(typeof(SyncScript));
        public static SyncScript Deserialize(FileInfo scriptFile)
        {
            using (StreamReader reader = scriptFile.OpenText())
            {
                SyncScript syncScript = (SyncScript)Serializer.Deserialize(reader);
                foreach (TableReflectingView syncTable in syncScript.SyncTables)
                {
                    syncTable.Parent = syncScript;
                }

                return syncScript;
            }
        }


        private SqlConnection m_SourceConnection;

        public SqlConnection SourceConnection
        {
            get {
                if (m_SourceConnection == null)
                {
                    m_SourceConnection = new SqlConnection(CrmSqlConnection);
                    m_SourceConnection.Open();
                }
                return m_SourceConnection;
            }
        }

        private SqlConnection m_TargetConnection;

        public SqlConnection TargetConnection
        {
            get
            {
                if (m_TargetConnection == null)
                {
                    m_TargetConnection = new SqlConnection(TargetSqlConnection);
                    m_TargetConnection.Open();
                }
                return m_TargetConnection;
            }
        }
	    
        public void Execute()
        {
            if (String.IsNullOrEmpty(CrmSqlConnection))
                throw new ArgumentNullException("CrmSqlConnection");

            // Create destination table reflecting source table / view
            if (SyncSchema)
            {
                foreach (TableReflectingView syncTable in SyncTables)
                {
                    syncTable.CopyTableSchema(true);
                }
            }

            // Create sprocs that copy view to table
            // CreateSyncSprocs(syncTables);

            
            CrmServer crmServer = new CrmServer();
            
            //foreach (TableReflectingView syncTable in SyncTables)
            //{
            //    syncTable.CopyTableSchema(false);
            //}

            foreach (TableReflectingView syncTable in SyncTables)
            {
                CrmSynchronizer synchronizer = new CrmSynchronizer(crmServer, syncTable, syncTable.Source.TableName.ToLowerInvariant());
                synchronizer.DoPreReplicationSync();
            }

            if (!String.IsNullOrEmpty(ReplicationPublicationName))
            {
                MergePushReplication replication = new MergePushReplication();
                replication.PublicationName = ReplicationPublicationName;

                replication.PerformMergeReplication(TargetConnection);
            }

            foreach (TableReflectingView syncTable in SyncTables)
            {
                CrmSynchronizer synchronizer = new CrmSynchronizer(crmServer, syncTable, syncTable.Source.TableName.ToLowerInvariant());
                synchronizer.DoPostReplicationSync();
            }
        }
    }
}
