using System;
using System.Collections.Generic;
using System.Text;
using System.Data;

using Microsoft.SqlServer.Management.Smo;
using Microsoft.SqlServer.Management.Smo.Agent;
using Microsoft.SqlServer.Management.Common;
using System.Data.SqlClient;
using System.Diagnostics;
using System.Collections.Specialized;
using System.Xml;
using System.Xml.Serialization;


namespace FastSqlSync
{
    /// <summary>
    /// Describes a table that holds the data in a CRM view
    /// </summary>
    public class TableReflectingView
    {
        private SyncScript m_Parent;

        /// <summary>
        /// The parent synchronization script
        /// </summary>
        public SyncScript Parent
        {
            get { return m_Parent; }
            set { m_Parent = value; }
        }
	
        private string m_ViewName;

        /// <summary>
        /// The name of the CRM view
        /// </summary>
        [XmlAttribute]
        public string ViewName
        {
            get { return m_ViewName; }
            set { m_ViewName = value; }
        }

        private string m_VersionColumn = "ModifiedOn";

        /// <summary>
        /// The column that stores the the version number
        /// </summary>
        [XmlAttribute]
        public string VersionColumn
        {
            get { return m_VersionColumn; }
            set { m_VersionColumn = value; }
        }

        private string m_KeyColumn;

        /// <summary>
        /// The column that holds the primary key
        /// </summary>
        [XmlAttribute]
        public string KeyColumn
        {
            get
            {
                if (String.IsNullOrEmpty(m_KeyColumn))
                {
                    return ViewName + "Id";
                }
                return m_KeyColumn;
            }
            set { m_KeyColumn = value; }
        }

        private SyncTableSpecification m_Source;
        public SyncTableSpecification Source
        {
            get
            {
                if (m_Source == null)
                {
                    m_Source = new SyncTableSpecification(Parent.SourceConnection, ViewName, KeyColumn, VersionColumn);
                }

                return m_Source; 
            }
        }

        private SyncTableSpecification m_Target;
        public SyncTableSpecification Target
        {
            get
            {
                if (m_Target == null)
                {
                    m_Target = new SyncTableSpecification(Parent.TargetConnection, ViewName, KeyColumn, VersionColumn);
                }

                return m_Target;
            }
        }

        public TableReflectingView()
        {
        }

        private Table GetSmoTable(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            Table table = database.Tables[connection.TableName];
            return table;
        }

        private View GetSmoView(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            View view = database.Views[connection.TableName];
            return view;
        }

        private TableViewBase GetSmoTableOrView(SyncTableSpecification connection)
        {
            Table table = GetSmoTable(connection);
            if (table != null) return table;

            View view = GetSmoView(connection);
            if (view != null) return view;

            throw new ArgumentOutOfRangeException();
        }

        private Table CreateSmoTable(SyncTableSpecification connection)
        {
            Database database = connection.SmoDatabase;
            Table table = new Table(database, connection.TableName);
            return table;
        }

        private string m_UpdateChangedRowsFromCrm;
        public string UpdateChangedRowsFromCrm
        {
            get
            {
                if (m_UpdateChangedRowsFromCrm == null) BuildSqlExpressions();
                return m_UpdateChangedRowsFromCrm;
            }
        }

        private string m_UpdateSingleRowFromCrm;
        public string UpdateSingleRowFromCrm
        {
            get
            {
                if (m_UpdateSingleRowFromCrm == null) BuildSqlExpressions();
                return m_UpdateSingleRowFromCrm;
            }
        }

        private string m_InsertNewRowsFromCrm;
        public string InsertNewRowsFromCrm
        {
            get
            {
                if (m_InsertNewRowsFromCrm == null) BuildSqlExpressions();
                return m_InsertNewRowsFromCrm;
            }
        }
        private string m_SelectRowsInsertedOnWeb;
        public string SelectRowsInsertedOnWeb
        {
            get
            {
                if (m_SelectRowsInsertedOnWeb == null) BuildSqlExpressions();
                return m_SelectRowsInsertedOnWeb;
            }
        }

        private string m_SelectRowsUpdatedOnWeb;
        public string SelectRowsUpdatedOnWeb
        {
            get
            {
                if (m_SelectRowsUpdatedOnWeb == null) BuildSqlExpressions();
                return m_SelectRowsUpdatedOnWeb;
            }
        }

        private string m_DeleteRowsRemovedInCrm;
        public string DeleteRowsRemovedInCrm
        {
            get
            {
                if (m_DeleteRowsRemovedInCrm == null) BuildSqlExpressions();
                return m_DeleteRowsRemovedInCrm;
            }
        }

        private void BuildSqlExpressions()
        {
            TableViewBase sourceTable = GetSmoTableOrView(Source);
            Table targetTable = GetSmoTable(Target);

            if (targetTable == null)
            {
                throw new InvalidOperationException();
            }

            ProgressLog.Current.WriteInfo(targetTable.Name);

            StringBuilder columnListBuilder = new StringBuilder();
            StringBuilder updateListBuilder = new StringBuilder();
            StringBuilder whereDifferentBuilder = new StringBuilder();

            string sourceTableName = Source.Connection.Database + ".." + sourceTable.Name;
            string targetTableName = Target.Connection.Database + ".." + targetTable.Name;

            foreach (Column sourceColumn in sourceTable.Columns)
            {
                string columnName = sourceColumn.Name;
                columnListBuilder.Append(", [" + columnName + "]");

                whereDifferentBuilder.Append("OR " + targetTableName + ".[" + columnName + "]<>src.[" + columnName + "]\n");

                if (columnName != Source.KeyColumn)
                    updateListBuilder.Append(", [" + columnName + "]=src.[" + columnName + "]\n");
            }

            string columnList = columnListBuilder.ToString().Substring(2);
            string updateList = updateListBuilder.ToString().Substring(2);
            string whereDifferent = whereDifferentBuilder.ToString().Substring(3);


            string insertNewRowsFromCrm =
                "INSERT INTO " + targetTableName +
                " (" + FieldMap.SpecialColumnChangedByWeb + ", " + FieldMap.SpecialColumnExistsInCrm + ", " + columnList + ") " +
                " SELECT 0, 1, " + columnList + " FROM " + sourceTableName +
                " WHERE " + Source.KeyColumn + " NOT IN " +
                " (SELECT " + Target.KeyColumn + " FROM " + targetTableName + ")\n";

            string deleteRowsRemovedInCrm = String.Format("DELETE FROM {0} WHERE ({1} NOT IN (SELECT {2} FROM {3})) AND ({4} = 1)\n",
                targetTableName, Target.KeyColumn, Source.KeyColumn, sourceTableName, FieldMap.SpecialColumnExistsInCrm);

            string updateSql;
            if (Source.VersionColumn != null)
            {
                updateSql =
                    "UPDATE " + targetTableName +
                    " SET " + FieldMap.SpecialColumnExistsInCrm + "=1, " + FieldMap.SpecialColumnChangedByWeb + "=0, " + updateList +
                    " FROM " + sourceTableName + " AS src " +
                    " WHERE (" + targetTableName + "." + Target.VersionColumn + "<>src." + Source.VersionColumn + ") " +
                    " AND " + targetTableName + "." + Target.KeyColumn + "=src." + Source.KeyColumn + "\n";

                m_UpdateSingleRowFromCrm =
                        "UPDATE " + targetTableName +
                        " SET " + FieldMap.SpecialColumnExistsInCrm + "=1, " + FieldMap.SpecialColumnChangedByWeb + "=0, " + updateList +
                        " FROM " + sourceTableName + " AS src " +
                        " WHERE (src." + Source.KeyColumn + "=@crmId) " +
                        " AND " + targetTableName + "." + Target.KeyColumn + "=src." + Source.KeyColumn + "\n";

            }
            else
            {
                throw new NotImplementedException();

                //updateSql = String.Format("UPDATE {0} SET {1} FROM {2} AS src WHERE ({3}) AND {4}.{5}=src.{6}\n",
                //    targetTableName, updateList, sourceTableName, whereDifferent, targetTableName, Target.KeyColumn, Target.KeyColumn);
            }

            string selectRowsInsertedOnWeb = String.Format("SELECT * FROM {0} WHERE ({1} NOT IN (SELECT {2} FROM {3})) AND (ISNULL({4},0) = 0)\n",
            targetTableName, Target.KeyColumn, Source.KeyColumn, sourceTableName, FieldMap.SpecialColumnExistsInCrm);

            string selectRowsUpdatedOnWeb = String.Format("SELECT * FROM {0} AS s JOIN {1} AS t ON (s.{4}=t.{5}) AND (ISNULL(s.{6},1) = 1)\n",
                Source.TableName, Target.TableName, Source.KeyColumn, Target.KeyColumn, Source.VersionColumn, Target.VersionColumn, FieldMap.SpecialColumnChangedByWeb);


            m_UpdateChangedRowsFromCrm = updateSql;
            m_InsertNewRowsFromCrm = insertNewRowsFromCrm;
            m_DeleteRowsRemovedInCrm = deleteRowsRemovedInCrm;
            m_SelectRowsInsertedOnWeb = selectRowsInsertedOnWeb;
            m_SelectRowsUpdatedOnWeb = selectRowsUpdatedOnWeb;
        }


        public static void CreateSyncAll(List<TableReflectingView> tables)
        {
            StringBuilder callBuilder = new StringBuilder();
            foreach (TableReflectingView tableReflector in tables)
            {
                callBuilder.Append("Exec " + tableReflector.SyncSprocName + "\n");
            }
            tables[0].CreateSproc("spWorldSync", callBuilder.ToString());
        }

        private void CreateSproc(string sprocName, string body)
        {
            StoredProcedure storedProcedure = Target.SmoDatabase.StoredProcedures[sprocName];
            if (storedProcedure == null)
            {
                storedProcedure = new StoredProcedure(Target.SmoDatabase, sprocName);
            }
            storedProcedure.TextMode = false;
            storedProcedure.AnsiNullsStatus = false;
            storedProcedure.QuotedIdentifierStatus = false;

            storedProcedure.TextBody = body;

            if (storedProcedure.State == SqlSmoState.Creating)
            {
                storedProcedure.Create();
            }
            else
            {
                storedProcedure.Alter();
            }
        }


        public string SyncSprocName
        {
            get
            {
                return ("spSync_" + Target.TableName);
            }
        }

        const int SpecialColumnCount = 2;

        public void CopyTableSchema(bool forceCreate)
        {
            Source.SmoServer.SetDefaultInitFields(typeof(Column), true);
            Target.SmoServer.SetDefaultInitFields(typeof(Column), true);

            TableViewBase sourceTable = GetSmoTableOrView(Source);
            Table targetTable = GetSmoTable(Target);
            if ((targetTable != null) && (!forceCreate))
            {
                // The table already exists, and we're not forcing creation
                // Do some superficial checks for changes

                // Number of columns must match
                if (targetTable.Columns.Count == (SpecialColumnCount + sourceTable.Columns.Count))
                {
                    return;
                }
            }

            if (targetTable == null)
            {
                targetTable = CreateSmoTable(Target);
            }

            ProgressLog.Current.WriteInfo(targetTable.Name);

            //if (sourceTable.Columns.Count == targetTable.Columns.Count)
            //{
            //    ProgressLog.Current.WriteInfo("\tSame number of columns - skipped");
            //    return;
            //}

            bool isNewTable = (targetTable.State == SqlSmoState.Creating);

            sourceTable.Columns.Refresh(true);
            targetTable.Columns.Refresh(true);

            foreach (Column sourceColumn in sourceTable.Columns)
            {
                bool isNewColumn = false;

                Column targetColumn = targetTable.Columns[sourceColumn.Name];
                if (targetColumn == null)
                {
                    targetColumn = new Column(targetTable, sourceColumn.Name);
                    isNewColumn = true;
                    DataType targetColumnType = sourceColumn.DataType;

                    if (targetColumnType.SqlDataType == SqlDataType.Timestamp)
                    {
                        if (sourceColumn.Nullable)
                        {
                            targetColumnType.SqlDataType = SqlDataType.Binary;
                            targetColumnType.MaximumLength = 8;
                        }
                        else
                        {
                            targetColumnType.SqlDataType = SqlDataType.VarBinary;
                            targetColumnType.MaximumLength = 8;
                        }
                    }

                    targetColumn.DataType = targetColumnType;
                }

                ProgressLog.Current.WriteInfo("\t" + sourceColumn.Name);
                //targetColumn.AnsiPaddingStatus = sourceColumn.AnsiPaddingStatus;

                if (isNewColumn || (targetColumn.RowGuidCol != sourceColumn.RowGuidCol))
                    targetColumn.RowGuidCol = sourceColumn.RowGuidCol;

                if (isNewColumn || (targetColumn.Collation != sourceColumn.Collation))
                    targetColumn.Collation = sourceColumn.Collation;
                if (isNewColumn || (targetColumn.Computed != sourceColumn.Computed))
                    targetColumn.Computed = sourceColumn.Computed;
                if (targetColumn.Computed)
                {
                    if (isNewColumn || (targetColumn.ComputedText != sourceColumn.ComputedText))
                        targetColumn.ComputedText = sourceColumn.ComputedText;
                    if (isNewColumn || (targetColumn.IsPersisted != sourceColumn.IsPersisted))
                        targetColumn.IsPersisted = sourceColumn.IsPersisted;
                }

                if (isNewColumn || (targetColumn.Default != sourceColumn.Default))
                    targetColumn.Default = sourceColumn.Default;
                //targetColumn.DefaultConstraint = sourceColumn.DefaultConstraint;
                if (isNewColumn || (targetColumn.DefaultSchema != sourceColumn.DefaultSchema))
                    targetColumn.DefaultSchema = sourceColumn.DefaultSchema;
                if (isNewColumn || (targetColumn.Identity != sourceColumn.Identity))
                    targetColumn.Identity = sourceColumn.Identity;
                if (isNewColumn || (targetColumn.IdentityIncrement != sourceColumn.IdentityIncrement))
                    targetColumn.IdentityIncrement = sourceColumn.IdentityIncrement;
                if (isNewColumn || (targetColumn.IdentitySeed != sourceColumn.IdentitySeed))
                    targetColumn.IdentitySeed = sourceColumn.IdentitySeed;
                //targetColumn.IsFullTextIndexed = sourceColumn.IsFullTextIndexed;
                if (isNewColumn || (targetColumn.NotForReplication != sourceColumn.NotForReplication))
                    targetColumn.NotForReplication = sourceColumn.NotForReplication;
                if (isNewColumn || (targetColumn.Nullable != sourceColumn.Nullable))
                    targetColumn.Nullable = sourceColumn.Nullable;
                if (isNewColumn || (targetColumn.RowGuidCol != sourceColumn.RowGuidCol))
                    targetColumn.RowGuidCol = sourceColumn.RowGuidCol;
                if (isNewColumn || (targetColumn.Rule != sourceColumn.Rule))
                    targetColumn.Rule = sourceColumn.Rule;
                if (isNewColumn || (targetColumn.RuleSchema != sourceColumn.RuleSchema))
                    targetColumn.RuleSchema = sourceColumn.RuleSchema;

                //foreach (Property sourceProperty in sourceColumn.Properties)
                //{
                //    if (!sourceProperty.Writable) continue;

                //    targetColumn.Properties[sourceProperty.Name].Value = sourceProperty.Value;
                //}

                switch (targetColumn.State)
                {
                    case SqlSmoState.Creating:
                        targetTable.Columns.Add(targetColumn);
                        break;
                    case SqlSmoState.Existing:
                        // No action needed
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }

            AddExtraColumn(targetTable, FieldMap.SpecialColumnExistsInCrm, DataType.Bit, "('0')");
            AddExtraColumn(targetTable, FieldMap.SpecialColumnChangedByWeb, DataType.Bit, "('0')");

            if (targetTable.State == SqlSmoState.Creating)
            {
                targetTable.Create();
            }
            else
            {
                targetTable.Alter();
            }


            //foreach (Index sourceIndex in sourceTable.Indexes)
            //{
            //    Index targetIndex = new Index(targetTable, sourceIndex.Name);
            //    foreach (IndexedColumn sourceIndexColumn in sourceIndex.IndexedColumns)
            //    {
            //        targetIndex.IndexedColumns.Add(new IndexedColumn(targetIndex, sourceIndexColumn.Name, sourceIndexColumn.Descending));
            //    }
            //    targetIndex.Create();
            //}

            string indexName = "PK_" + Source.KeyColumn;
            Index targetKeyIndex = targetTable.Indexes[indexName];
            if (targetKeyIndex != null)
            {
                // This seems to be the procedure for dropping a Primary Key
                targetKeyIndex.IndexKeyType = IndexKeyType.None;
                targetKeyIndex.MarkForDrop(true);
                targetTable.Alter();
                //   targetKeyIndex.Drop();
            }
            targetKeyIndex = new Index(targetTable, indexName);
            targetKeyIndex.IndexedColumns.Add(new IndexedColumn(targetKeyIndex, Source.KeyColumn));
            targetKeyIndex.IndexKeyType = IndexKeyType.DriPrimaryKey;
            targetKeyIndex.IsUnique = true;
            targetKeyIndex.Create();

            Column keyColumn = targetTable.Columns[Source.KeyColumn];
            keyColumn.RowGuidCol = true;

            keyColumn.Alter();

        }

        /// <summary>
        /// Creates a target column that is 'extra' - doesn't exist in the source
        /// </summary>
        /// <param name="targetTable"></param>
        /// <param name="columnName"></param>
        /// <param name="sqlDbType"></param>
        private void AddExtraColumn(Table targetTable, string columnName, DataType sqlDbType, string defaultValue)
        {
            Column targetColumn = targetTable.Columns[columnName];
            if (targetColumn == null)
            {
                targetColumn = new Column(targetTable, columnName, sqlDbType);
                targetTable.Columns.Add(targetColumn);
                if (targetColumn.State != SqlSmoState.Creating)
                    targetTable.Alter();
            }

            DefaultConstraint defaultConstraint = targetColumn.DefaultConstraint;
            if (defaultConstraint == null)
                defaultConstraint = targetColumn.AddDefaultConstraint();
            defaultConstraint.Text = defaultValue;
            if (targetColumn.State != SqlSmoState.Creating)
                targetColumn.Alter();
        }
    }
}
